@@ -6,20 +6,24 @@ parser = argparse.ArgumentParser(
     description="Start a local webserver with a Python terminal."
 )
 parser.add_argument(
-    "--port", type=int, default=8000, help="port for the http server to listen on"
+    "--port",
+    type=int,
+    default=8000,
+    help="port for the http server to listen on",
 )
 parser.add_argument(
-    "--bind", type=str, default="127.0.0.1", help="Bind address (empty for all)"
+    "--bind",
+    type=str,
+    default="127.0.0.1",
+    help="Bind address (empty for all)",
 )
 
 
 class MyHTTPRequestHandler(server.SimpleHTTPRequestHandler):
     extensions_map = server.SimpleHTTPRequestHandler.extensions_map.copy()
-    extensions_map.update(
-        {
-            ".wasm": "application/wasm",
-        }
-    )
+    extensions_map.update({
+        ".wasm": "application/wasm",
+    })
 
     def end_headers(self) -> None:
         self.send_my_headers()
@@ -42,5 +46,6 @@ def main() -> None:
         bind=args.bind,
     )
 
+
 if __name__ == "__main__":
     main()
